@@ -1,8 +1,26 @@
 # -*- coding: utf-8 -*-
 #
 # Picard, the next-generation MusicBrainz tagger
+#
 # Copyright (C) 2004 Robert Kaye
-# Copyright (C) 2006-2007 Lukáš Lalinský
+# Copyright (C) 2006-2009, 2011-2012, 2014 Lukáš Lalinský
+# Copyright (C) 2008 Gary van der Merwe
+# Copyright (C) 2008 Hendrik van Antwerpen
+# Copyright (C) 2008 ojnkpjg
+# Copyright (C) 2008-2011, 2014, 2018-2020 Philipp Wolfer
+# Copyright (C) 2009 Nikolai Prokoschenko
+# Copyright (C) 2011-2012 Chad Wilson
+# Copyright (C) 2011-2013, 2019 Michael Wiencek
+# Copyright (C) 2012-2013, 2016-2017 Wieland Hoffmann
+# Copyright (C) 2013, 2018 Calvin Walton
+# Copyright (C) 2013-2015, 2017 Sophist-UK
+# Copyright (C) 2013-2015, 2017-2019 Laurent Monin
+# Copyright (C) 2016 Suhas
+# Copyright (C) 2016-2018 Sambhav Kothari
+# Copyright (C) 2017 Antonio Larrosa
+# Copyright (C) 2018 Vishal Choudhary
+# Copyright (C) 2019 Joel Lintunen
+# Copyright (C) 2020 Gabriel Ferreira
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -18,6 +36,7 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
+
 from collections import (
     OrderedDict,
     defaultdict,
@@ -64,6 +83,7 @@ from picard.util import (
     find_best_match,
     format_time,
     mbid_validate,
+    process_events_iter,
 )
 from picard.util.imagelist import (
     add_metadata_images,
@@ -151,6 +171,13 @@ class Album(DataObject, Item):
                     file.metadata['musicbrainz_discid'] = track_discids
                     file.update()
 
+    def get_next_track(self, track):
+        try:
+            index = self.tracks.index(track)
+            return self.tracks[index + 1]
+        except (IndexError, ValueError):
+            return None
+
     def get_album_artists(self):
         """Returns the list of album artists (as AlbumArtist objects)"""
         return self._album_artists
@@ -267,6 +294,11 @@ class Album(DataObject, Item):
             del self._new_metadata
             del self._new_tracks
             self.update()
+            if not self._requests:
+                self.loaded = True
+                for func, always in self._after_load_callbacks:
+                    if always:
+                        func()
             return
 
         if self._requests > 0:
@@ -287,7 +319,6 @@ class Album(DataObject, Item):
                 mm = Metadata()
                 mm.copy(self._new_metadata)
                 medium_to_metadata(medium_node, mm)
-                discpregap = False
                 format = medium_node.get('format')
                 if format:
                     all_media.append(format)
@@ -295,29 +326,43 @@ class Album(DataObject, Item):
                 for dj in djmix_ars.get(mm["discnumber"], []):
                     mm.add("djmixer", dj)
 
+                if va:
+                    mm["compilation"] = "1"
+                else:
+                    del mm["compilation"]
+
                 if 'discs' in medium_node:
                     discids = [disc.get('id') for disc in medium_node['discs']]
                     mm['~musicbrainz_discids'] = discids
                     mm['musicbrainz_discid'] = list(self._discids.intersection(discids))
 
                 if "pregap" in medium_node:
-                    discpregap = True
                     absolutetracknumber += 1
-                    track = self._finalize_loading_track(medium_node['pregap'], mm, artists, va, absolutetracknumber, discpregap)
-                    track.metadata['~pregap'] = "1"
+                    mm['~discpregap'] = '1'
+                    extra_metadata = {
+                        '~pregap': '1',
+                        '~absolutetracknumber': absolutetracknumber,
+                    }
+                    self._finalize_loading_track(medium_node['pregap'], mm, artists, extra_metadata)
 
                 track_count = medium_node['track-count']
                 if track_count:
                     tracklist_node = medium_node['tracks']
                     for track_node in tracklist_node:
                         absolutetracknumber += 1
-                        track = self._finalize_loading_track(track_node, mm, artists, va, absolutetracknumber, discpregap)
+                        extra_metadata = {
+                            '~absolutetracknumber': absolutetracknumber,
+                        }
+                        self._finalize_loading_track(track_node, mm, artists, extra_metadata)
 
                 if "data-tracks" in medium_node:
                     for track_node in medium_node['data-tracks']:
                         absolutetracknumber += 1
-                        track = self._finalize_loading_track(track_node, mm, artists, va, absolutetracknumber, discpregap)
-                        track.metadata['~datatrack'] = "1"
+                        extra_metadata = {
+                            '~datatrack': '1',
+                            '~absolutetracknumber': absolutetracknumber,
+                        }
+                        self._finalize_loading_track(track_node, mm, artists, extra_metadata)
 
             totalalbumtracks = absolutetracknumber
             self._new_metadata['~totalalbumtracks'] = totalalbumtracks
@@ -334,6 +379,9 @@ class Album(DataObject, Item):
 
         if not self._requests:
             self.enable_update_metadata_images(False)
+            for track in self._new_tracks:
+                track.orig_metadata.copy(track.metadata)
+
             # Prepare parser for user's script
             for s_name, s_text in enabled_tagger_scripts_texts():
                 parser = ScriptParser()
@@ -344,6 +392,7 @@ class Album(DataObject, Item):
                     except ScriptError:
                         log.exception("Failed to run tagger script %s on track", s_name)
                     track.metadata.strip_whitespace()
+                    track.scripted_metadata.update(track.metadata)
                 # Run tagger script for the album itself
                 try:
                     parser.eval(s_text, self._new_metadata)
@@ -356,6 +405,7 @@ class Album(DataObject, Item):
                 for file in list(track.linked_files):
                     file.move(self.unmatched_files)
             self.metadata = self._new_metadata
+            self.orig_metadata.copy(self.metadata)
             self.tracks = self._new_tracks
             del self._new_metadata
             del self._new_tracks
@@ -373,11 +423,13 @@ class Album(DataObject, Item):
                 },
                 timeout=3000
             )
-            for func in self._after_load_callbacks:
+            for func, always in self._after_load_callbacks:
                 func()
             self._after_load_callbacks = []
+            if self.item.isSelected():
+                self.tagger.window.refresh_metadatabox()
 
-    def _finalize_loading_track(self, track_node, metadata, artists, va, absolutetracknumber, discpregap):
+    def _finalize_loading_track(self, track_node, metadata, artists, extra_metadata=None):
         # As noted in `_parse_release` above, the release artist nodes
         # may contain supplementary data that isn't present in track
         # artist nodes. Similarly, the track artists may contain
@@ -394,18 +446,12 @@ class Album(DataObject, Item):
         tm = track.metadata
         tm.copy(metadata)
         track_to_metadata(track_node, track)
-        tm["~absolutetracknumber"] = absolutetracknumber
-        track.orig_metadata.copy(tm)
         track._customize_metadata()
 
         self._new_metadata.length += tm.length
         artists.add(tm["artist"])
-        if va:
-            tm["compilation"] = "1"
-        else:
-            del tm["compilation"]
-        if discpregap:
-            tm["~discpregap"] = "1"
+        if extra_metadata:
+            tm.update(extra_metadata)
 
         # Run track metadata plugins
         try:
@@ -453,11 +499,11 @@ class Album(DataObject, Item):
             self.id, self._release_request_finished, inc=inc,
             mblogin=require_authentication, priority=priority, refresh=refresh)
 
-    def run_when_loaded(self, func):
+    def run_when_loaded(self, func, always=False):
         if self.loaded:
             func()
         else:
-            self._after_load_callbacks.append(func)
+            self._after_load_callbacks.append((func, always))
 
     def stop_loading(self):
         if self.load_task:
@@ -480,7 +526,7 @@ class Album(DataObject, Item):
         file.metadata_images_changed.disconnect(self.update_metadata_images)
         remove_metadata_images(self, [file])
 
-    def _match_files(self, files, recordingid=None, threshold=0):
+    def _match_files(self, files, threshold=0):
         """Match files to tracks on this album, based on metadata similarity or recordingid."""
         tracks_cache = defaultdict(lambda: None)
 
@@ -500,8 +546,8 @@ class Album(DataObject, Item):
         for file in list(files):
             if file.state == File.REMOVED:
                 continue
-            # if we have a recordingid to match against, use that in priority
-            recid = recordingid or file.metadata['musicbrainz_recordingid']
+            # if we have a recordingid to match against, use that in priority
+            recid = file.match_recordingid or file.metadata['musicbrainz_recordingid']
             if recid and mbid_validate(recid):
                 if not tracks_cache:
                     build_tracks_cache()
@@ -516,12 +562,11 @@ class Album(DataObject, Item):
 
             # try to match by similarity
             def candidates():
-                for track in self.tracks:
+                for track in process_events_iter(self.tracks):
                     yield SimMatchAlbum(
                         similarity=track.metadata.compare(file.orig_metadata),
                         track=track
                     )
-                    QtCore.QCoreApplication.processEvents()
 
             no_match = SimMatchAlbum(similarity=-1, track=self.unmatched_files)
             best_match = find_best_match(candidates, no_match)
@@ -531,13 +576,15 @@ class Album(DataObject, Item):
             else:
                 yield (file, best_match.result.track)
 
-    def match_files(self, files, recordingid=None):
+    def match_files(self, files):
         """Match and move files to tracks on this album, based on metadata similarity or recordingid."""
-        moves = self._match_files(files,
-                                  recordingid=recordingid,
-                                  threshold=config.setting['track_matching_threshold'])
-        for file, target in moves:
-            file.move(target)
+        if self.loaded:
+            moves = self._match_files(files, threshold=config.setting['track_matching_threshold'])
+            for file, target in moves:
+                file.move(target)
+        else:
+            for file in files:
+                file.move(self.unmatched_files)
 
     def can_save(self):
         return self._files > 0
@@ -558,7 +605,7 @@ class Album(DataObject, Item):
         return True
 
     def can_view_info(self):
-        return (self.loaded and (self.metadata.images or self.orig_metadata.images)) or self.errors
+        return self.loaded or self.errors
 
     def is_album_like(self):
         return True
