@@ -1,7 +1,10 @@
 # -*- coding: utf-8 -*-
 #
 # Picard, the next-generation MusicBrainz tagger
+#
 # Copyright (C) 2017 Sambhav Kothari
+# Copyright (C) 2018 Laurent Monin
+# Copyright (C) 2018-2019 Philipp Wolfer
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -17,6 +20,7 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
+
 import re
 
 from PyQt5.QtCore import QUrl
@@ -152,7 +156,8 @@ class MBAPIHelper(APIHelper):
 
         is_search = kwargs.pop("search", False)
         if is_search:
-            if config.setting["use_adv_search_syntax"]:
+            use_advanced_search = kwargs.pop("advanced_search", config.setting["use_adv_search_syntax"])
+            if use_advanced_search:
                 query = kwargs["query"]
             else:
                 query = escape_lucene_query(kwargs["query"]).strip().lower()
